import torch 
import torch.nn as nn 
from .cells import * 
from ..base import RNN_Base
from .constants import RNNMAP


class SnapShot_Base(RNN_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        cell_type: str, 
        sample_rate: int,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(sample_rate)

        # arguments 
        self.inp_channel = inp_channel
        self.out_channel = out_channel
        self.rnn_size = rnn_size
        self.cell_type = cell_type
        self.layer_norm = layer_norm
        self.rnn_bias = rnn_bias
        
        if self.layer_norm:
            self.main_rnn = LearnableCell(
                input_size = self.inp_channel, 
                hidden_size = self.rnn_size, 
                cell_type = self.cell_type,
                layer_norm = self.layer_norm,
                bias = self.rnn_bias
            )
        else:
            if self.cell_type == 'gru':
                self.main_rnn = torch.nn.GRU(
                    input_size = self.inp_channel, 
                    hidden_size = self.rnn_size, 
                    batch_first = True,
                    bias = self.rnn_bias
                )
            elif self.cell_type == 'lstm':
                self.main_rnn = torch.nn.LSTM(
                    input_size = self.inp_channel, 
                    hidden_size = self.rnn_size, 
                    batch_first = True,
                    bias = self.rnn_bias
                )
            elif self.cell_type == 'vanilla_rnn':
                self.main_rnn = torch.nn.RNN(
                    input_size = self.inp_channel, 
                    hidden_size = self.rnn_size, 
                    batch_first = True,
                    bias = self.rnn_bias
                )
        
        self.linear_out = nn.Linear(
            self.rnn_size,
            self.out_channel,
            bias = self.rnn_bias
        )

    def forward(self, x, c, h0):

        
        # B x C x T -> B x T x C
        x = x.permute(0, 2, 1)

        # for gru
        x, h = self.main_rnn(x, h0)
        x = self.linear_out(x)

        # out
        x = x.permute(0, 2, 1)

        return x, h, None


class Concat_Base(RNN_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        cell_type: str, 
        sample_rate: int,
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(sample_rate)

        # arguments 
        self.inp_channel = inp_channel
        self.out_channel = out_channel
        self.rnn_size = rnn_size
        self.cell_type = cell_type
        self.num_conds = num_conds
        self.layer_norm = layer_norm
        self.rnn_bias = rnn_bias
        
        if self.layer_norm:
            self.main_rnn = LearnableCell(
                input_size = self.inp_channel + self.num_conds, 
                hidden_size = self.rnn_size, 
                cell_type = self.cell_type,
                layer_norm = self.layer_norm,
                bias = self.rnn_bias
            )
        else:
            if self.cell_type == 'gru':
                self.main_rnn = torch.nn.GRU(
                    input_size = self.inp_channel + self.num_conds, 
                    hidden_size = self.rnn_size, 
                    batch_first = True,
                    bias = self.rnn_bias
                )
            elif self.cell_type == 'lstm':
                self.main_rnn = torch.nn.LSTM(
                    input_size = self.inp_channel + self.num_conds, 
                    hidden_size = self.rnn_size, 
                    batch_first = True,
                    bias = self.rnn_bias
                )
            elif self.cell_type == 'vanilla_rnn':
                self.main_rnn = torch.nn.RNN(
                    input_size = self.inp_channel + self.num_conds, 
                    hidden_size = self.rnn_size, 
                    batch_first = True,
                    bias = self.rnn_bias
                )
        
        self.linear_out = nn.Linear(
            self.rnn_size,
            self.out_channel,
            bias = self.rnn_bias
        )

    def forward(self, x, c, h0):

        if len(c.shape) == 2:
            c = c.unsqueeze(-1)
            c = c.expand(-1, -1, x.shape[-1])

        # cat x and cond
        if self.num_conds > 0:
            x = torch.cat((x, c), dim=1)
        
        # B x C x T -> B x T x C
        x = x.permute(0, 2, 1)

        # for gru
        x, h = self.main_rnn(x, h0)
        x = self.linear_out(x)

        # out
        x = x.permute(0, 2, 1)

        return x, h, None

class FiLM_Base(RNN_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        cell_type: str, 
        sample_rate: int,
        num_conds: int = 0,
        mlp_size: int = 16,
        n_mlp_blocks: int = 2,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
        filmed: bool = True
    ):
        super().__init__(sample_rate)

        if num_conds <= 0:
            raise ValueError(f'FiLM networks is used only when condition is provided')
        
        # arguments 
        self.inp_channel = inp_channel
        self.out_channel = out_channel
        self.rnn_size = rnn_size
        self.n_mlp_blocks = n_mlp_blocks
        self.mlp_size = mlp_size
        self.num_conds = num_conds
        self.layer_norm = layer_norm
        self.rnn_bias = rnn_bias
        self.cell_type = cell_type

        # MLP 
        self.SIZE_MAP = RNNMAP 

        self.gen = nn.ModuleList()
        for n in range(n_mlp_blocks):
            _input_features = num_conds if n == 0 else mlp_size
            self.gen.append(nn.Sequential(
                nn.Linear(_input_features, mlp_size, bias=True),
                nn.LeakyReLU(0.1)
            ))
        
        if n_mlp_blocks == 0:
            self.gen.append(nn.Linear(num_conds, rnn_size*(self.SIZE_MAP[self.cell_type]*4), bias=True))
        else:
            self.gen.append(nn.Linear(mlp_size, rnn_size*(self.SIZE_MAP[self.cell_type]*4), bias=True))
        
        
        self.gen = torch.nn.Sequential(*self.gen)

        self.main_rnn = LearnableCell(
            input_size = self.inp_channel, 
            hidden_size = self.rnn_size, 
            cell_type = self.cell_type,
            layer_norm = self.layer_norm,
            bias = self.rnn_bias,
            filmed = filmed
        )

        self.linear_out = nn.Linear(
            self.rnn_size,
            self.out_channel,
            bias = self.rnn_bias
        )
        
    def forward(self, x, c, h0):

        ih_ab, hh_ab = torch.chunk(self.gen(c), 2, dim=-1)
        a_i, b_i = torch.chunk(ih_ab, 2, dim=-1)
        a_h, b_h = torch.chunk(hh_ab, 2, dim=-1)

        
        x = x.permute(0, 2, 1)

        x, h = self.main_rnn(
            x, 
            h0,
            a_i,
            b_i,
            a_h,
            b_h
        )
    
        x = self.linear_out(x)
        
        x = x.permute(0, 2, 1)

        return x, h, None

class StaticHyper_Base(RNN_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        cell_type: str, 
        sample_rate: int,
        n_mlp_blocks: int = 3,
        mlp_size: int = 8,
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(sample_rate)

        if num_conds <= 0:
            raise ValueError(f'Static hyper networks is used only when condition is provided')
        
        # arguments 
        self.inp_channel = inp_channel
        self.out_channel = out_channel
        self.rnn_size = rnn_size
        self.n_mlp_blocks = n_mlp_blocks
        self.mlp_size = mlp_size
        self.num_conds = num_conds
        self.layer_norm = layer_norm
        self.rnn_bias = rnn_bias
        self.cell_type = cell_type
        

        # cells 
        self.main_rnn = NonlearnableCell(
            hidden_size = self.rnn_size, 
            cell_type = self.cell_type,
            layer_norm = self.layer_norm, 
        )

        self.linear_out = nn.Linear(
            self.rnn_size,
            self.out_channel,
            bias = self.rnn_bias
        )
        
        # MLP 
        self.SIZE_MAP = RNNMAP 
            

        if self.num_conds > 0: # if condition
            self.cond_mlp = nn.ModuleList()
            for n in range(self.n_mlp_blocks-1):
                _input_features = self.num_conds if n == 0 else self.mlp_size
                self.cond_mlp.append(nn.Sequential(
                    nn.Linear(_input_features, self.mlp_size, bias=True),
                    nn.LeakyReLU(0.1)
                ))
            self.cond_mlp = torch.nn.Sequential(*self.cond_mlp)

            # initialization 
            for idx, w in enumerate(self.cond_mlp.parameters()):
                if idx == 0:
                    bound = math.sqrt(self.SIZE_MAP[self.cell_type] / (self.mlp_size * self.num_conds))
                else:
                    bound = math.sqrt(self.SIZE_MAP[self.cell_type] / (self.mlp_size * self.mlp_size))
                w.data.uniform_(-bound, bound)


            _proj_ih_out = self.SIZE_MAP[self.cell_type] * self.inp_channel * self.rnn_size
            _proj_hh_out = self.SIZE_MAP[self.cell_type] * self.rnn_size * self.rnn_size
            _bias_out = self.SIZE_MAP[self.cell_type] * self.rnn_size

            self.proj_ih = torch.nn.Linear(
                self.mlp_size, 
                _proj_ih_out, 
                bias=True
            )
            
            self.proj_hh = torch.nn.Linear(
                self.mlp_size, 
                _proj_hh_out,
                bias = True
            )


            if self.rnn_bias: 
                self.proj_bih = torch.nn.Linear(
                    self.mlp_size,
                    _bias_out,
                    bias = True 
                )

                self.proj_bhh = torch.nn.Linear(
                    self.mlp_size,
                    _bias_out,
                    bias = True
                )
        

    def forward(self, x, c, h0):

        w = self.cond_mlp(c) 
        w_ih = self.proj_ih(w).reshape(-1, self.inp_channel, self.rnn_size * self.SIZE_MAP[self.cell_type])
        w_hh = self.proj_hh(w).reshape(-1, self.rnn_size, self.rnn_size * self.SIZE_MAP[self.cell_type])
        b_ih = None
        b_hh = None

        if self.rnn_bias:
            b_ih = self.proj_bih(w).reshape(-1, 1, self.rnn_size * self.SIZE_MAP[self.cell_type])
            b_hh = self.proj_bhh(w).reshape(-1, 1, self.rnn_size * self.SIZE_MAP[self.cell_type])
        
        x = x.permute(0, 2, 1)

        # core rnn
        x, h_out = self.main_rnn(x, h0, w_ih, w_hh, b_ih, b_hh)
        
        x = self.linear_out(x)
        
        x = x.permute(0, 2, 1)

        

        if self.cell_type == 'lstm':
            __h, __c = h_out
            return x, (__h.detach(), __c.detach()), (w_ih.detach(), w_hh.detach(), b_ih.detach(), b_hh.detach())
        
        return x, h_out.detach(), (w_ih.detach(), w_hh.detach(), b_ih.detach(), b_hh.detach())

class DynamicHyper_Base(RNN_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        cell_type: str, 
        sample_rate: int,
        hyper_rnn_size: int = 8,
        n_z_size: int = 8,
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(sample_rate)   

        # arguments 
        self.inp_channel = inp_channel
        self.out_channel = out_channel
        self.rnn_size = rnn_size
        self.hyper_rnn_size = hyper_rnn_size
        self.n_z_size = n_z_size
        self.num_conds = num_conds
        self.layer_norm = layer_norm
        self.rnn_bias = rnn_bias
        self.cell_type = cell_type

        self.main_rnn = DynamicHyperCell(
            input_size = self.num_conds,
            hidden_size = self.rnn_size, 
            cell_type = self.cell_type,
            hyper_size = self.hyper_rnn_size, 
            n_z = self.n_z_size, 
            layer_norm = self.layer_norm, 
            bias = rnn_bias
        )

        self.linear_out = nn.Linear(
            self.rnn_size,
            self.out_channel,
            bias = self.rnn_bias
        )
    

    def forward(
        self,
        x,
        c,
        h0,
        h_hat0,
    ):
        # B x C x T -> B x T x C
        x = x.permute(0, 2, 1)

        
        # for gru
        x, h = self.main_rnn(x, c, h0, h_hat0)
        
        x = self.linear_out(x)

        # out
        x = x.permute(0, 2, 1)

        return x, h, None


import torch 
import torch.nn as nn 

from ..base import CNN_Base
from ..utils import * 
from .blocks import TCNBlock, AnalogTCNBlock


class SnapShotTCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = 0
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                TCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = 0,
                    kernel_size= kernel_size, 
                    dilation=dilation,
                    causal=causal,
                    filmed= False,
                    hypered = False,
                    bias = bias))

        self.output = torch.nn.Conv1d(out_ch, self.n_output, kernel_size=1)

    def forward(self, x, c, *args):
        
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x = block(x, None)
        
        # out
        out = self.output(x)
        
        return out

    def compute_receptive_field(self): # in samples 
        rf = self.kernel_size
        for n in range(1, self.n_blocks):
            dilation = self.dilation_growth ** n
            rf = rf + ((self.kernel_size-1) * dilation)
        return rf, rf/self.sample_rate * 1000 # samples, ms
    

    def compute_num_of_params(self):
        return (sum(p.numel() for p in self.parameters()), sum(p.numel() for p in self.parameters() if p.requires_grad))
    
class SnapShotAnalogTCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = 0
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                AnalogTCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = 0,
                    kernel_size= kernel_size, 
                    dilation=dilation,
                    causal=causal,
                    bias = bias))

        self.output = torch.nn.Conv1d(out_ch, self.n_output, kernel_size=1)

        self.prepare(self.sample_rate, self.kernel_size, 1)

    def forward(self, x, c, *args):
        
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x = block(x, None)
        
        # out
        out = self.output(x)
        
        return out

    def compute_receptive_field(self): # in samples 
        rf = self.kernel_size
        for n in range(1, self.n_blocks):
            dilation = self.dilation_growth ** n
            rf = rf + ((self.kernel_size-1) * dilation)
        return rf, rf/self.sample_rate * 1000 # samples, ms
    

    def compute_num_of_params(self):
        return (sum(p.numel() for p in self.parameters()), sum(p.numel() for p in self.parameters() if p.requires_grad))
    
    def prepare(self, sample_rate, kernel_size, stride):
        for b in self.blocks:
            b.prepare(sample_rate, kernel_size, stride)

class ConcatTCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = n_cond
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else (self.n_inp + self.n_cond)
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                TCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = 0,
                    kernel_size= kernel_size, 
                    dilation=dilation,
                    causal=causal,
                    filmed= False,
                    hypered = False,
                    bias = bias))

        self.output = torch.nn.Conv1d(out_ch, self.n_output, kernel_size=1)

    def forward(self, x, c, *args):

        if c is not None and len(c.shape) == 2:
            c = c.unsqueeze(-1)
            c = c.expand(-1, -1, x.shape[-1])

        # cat x and cond
        if self.n_cond > 0:
            x = torch.cat((x, c), dim=1)
        
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x = block(x, None)
        
        # out
        out = self.output(x)
        
        return out

    def compute_receptive_field(self): # in samples 
        rf = self.kernel_size
        for n in range(1, self.n_blocks):
            dilation = self.dilation_growth ** n
            rf = rf + ((self.kernel_size-1) * dilation)
        return rf, rf/self.sample_rate * 1000 # samples, ms
    

    def compute_num_of_params(self):
        return (sum(p.numel() for p in self.parameters()), sum(p.numel() for p in self.parameters() if p.requires_grad))
    
class FiLMTCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
        pre_film_size: int = 16,
        pre_film_blocks: int = 3,
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = n_cond
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.pre_film = None
        if self.n_cond > 0:
            pre_film = torch.nn.ModuleList()
            for n in range(pre_film_blocks-1):
                inp = pre_film_size if n > 0 else n_cond
                out = pre_film_size
                pre_film.append(nn.Sequential(
                    nn.Linear(inp, out),
                    nn.LeakyReLU(0.2)
                ))
            pre_film.append(nn.Linear(pre_film_size, pre_film_size))
            self.pre_film = nn.Sequential(*pre_film)

        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                TCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = pre_film_size,
                    kernel_size = kernel_size, 
                    dilation = dilation,
                    causal = causal,
                    filmed = True,
                    hypered = False,
                    bias = bias))

        self.output = torch.nn.Conv1d(out_ch, self.n_output, kernel_size=1)

    def forward(self, x, c, *args):
        
        if self.pre_film is not None and self.n_cond > 0:
            c = self.pre_film(c)
        
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x = block(x, c)
        
        # out
        out = self.output(x)
        
        return out

    def compute_receptive_field(self): # in samples 
        rf = self.kernel_size
        for n in range(1, self.n_blocks):
            dilation = self.dilation_growth ** n
            rf = rf + ((self.kernel_size-1) * dilation)
        return rf, rf/self.sample_rate * 1000 # samples, ms
    

    def compute_num_of_params(self):
        return (sum(p.numel() for p in self.parameters()), sum(p.numel() for p in self.parameters() if p.requires_grad))

class HyperTCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True, 
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = n_cond
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                TCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = n_cond,
                    kernel_size = kernel_size, 
                    dilation = dilation,
                    causal = causal,
                    filmed = False,
                    hypered = True,
                    bias = bias))

        self.output = torch.nn.Conv1d(out_ch, self.n_output, kernel_size=1)

    def forward(self, x, c, *args):
        
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x = block(x, c)
        
        # out
        out = self.output(x)
        
        return out

    def compute_receptive_field(self): # in samples 
        rf = self.kernel_size
        for n in range(1, self.n_blocks):
            dilation = self.dilation_growth ** n
            rf = rf + ((self.kernel_size-1) * dilation)
        return rf, rf/self.sample_rate * 1000 # samples, ms
    

    def compute_num_of_params(self):
        return (sum(p.numel() for p in self.parameters()), sum(p.numel() for p in self.parameters() if p.requires_grad))





# functions that implement analysis and synthesis of sounds using the Stochastic Model
# (for example usage check stochasticModel_function.py in the models_interface directory)

import numpy as np
from scipy.signal import resample
from scipy.signal.windows import hann
from scipy.interpolate import splrep, splev
from scipy.fft import fft, ifft
from . import utilFunctions as UF

def hertz_to_mel(f):
    """
    Conversion from hertz scale to mel scale
    """
    return 2595 * np.log10(1 + f / 700)

def mel_to_hetz(m):
    """
    Conversion from mel scale to hertz scale
    """
    return 700 * (10 ** (m / 2595) - 1)

def stochasticModelAnal(x, H, N, stocf, fs=44100, melScale=1):
    """
	Stochastic analysis of a sound
	x: input array sound, H: hop size, N: fftsize
	stocf: decimation factor of mag spectrum for stochastic analysis, bigger than 0, maximum of 1
	fs: sampling rate
	melScale: choose between linear scale, 0, or mel scale, 1
	returns stocEnv: stochastic envelope
	"""

    hN = N // 2 + 1  # positive size of fft
    No2 = N // 2  # half of N
    if (hN * stocf < 3):  # raise exception if decimation factor too small
        raise ValueError("Stochastic decimation factor too small")

    if (stocf > 1):  # raise exception if decimation factor too big
        raise ValueError("Stochastic decimation factor above 1")

    if (H <= 0):  # raise error if hop size 0 or negative
        raise ValueError("Hop size (H) smaller or equal to 0")

    if not (UF.isPower2(N)):  # raise error if N not a power of two
        raise ValueError("FFT size (N) is not a power of 2")

    w = hann(N)  # analysis window
    x = np.append(np.zeros(No2), x)  # add zeros at beginning to center first window at sample 0
    x = np.append(x, np.zeros(No2))  # add zeros at the end to analyze last sample
    pin = No2  # initialize sound pointer in middle of analysis window
    pend = x.size - No2  # last sample to start a frame
    if melScale == 1:
        binFreqsMel = hertz_to_mel(np.arange(hN)*fs/float(N))
        uniformMelFreq = np.linspace(binFreqsMel[0], binFreqsMel[-1], hN)
    while pin <= pend:
        xw = x[pin - No2:pin + No2] * w  # window the input sound
        X = fft(xw)  # compute FFT
        mX = 20 * np.log10(abs(X[:hN]))  # magnitude spectrum of positive frequencies
        if melScale == 1:
            spl = splrep(binFreqsMel, np.maximum(-200, mX))
            mY = resample(splev(uniformMelFreq, spl), int(stocf * hN))  # decimate the mag spectrum
        else:
            mY = resample(np.maximum(-200, mX), int(stocf * hN))  # decimate the mag spectrum
        if pin == No2:  # first frame
            stocEnv = np.array([mY])
        else:  # rest of frames
            stocEnv = np.vstack((stocEnv, np.array([mY])))
        pin += H  # advance sound pointer
    return stocEnv


def stochasticModelSynth(stocEnv, H, N, fs=44100, melScale=1):
    """
	Stochastic synthesis of a sound
	stocEnv: stochastic envelope; H: hop size; N: fft size
	fs: sampling rate
	melScale: choose between linear scale, 0, or mel scale, 1 (should match the analysis)
	returns y: output sound
	"""

    if not (UF.isPower2(N)):  # raise error if N not a power of two
        raise ValueError("N is not a power of two")

    hN = N // 2 + 1  # positive size of fft
    No2 = N // 2  # half of N
    L = stocEnv[:, 0].size  # number of frames
    ysize = H * (L + 3)  # output sound size
    y = np.zeros(ysize)  # initialize output array
    ws = 2 * hann(N)  # synthesis window
    pout = 0  # output sound pointer
    if melScale == 1:
        binFreqsMel = hertz_to_mel(np.arange(hN)*fs/float(N))
        uniformMelFreq = np.linspace(binFreqsMel[0], binFreqsMel[-1], hN)
    for l in range(L):
        mY = resample(stocEnv[l, :], hN)  # interpolate to original size
        if melScale == 1:
            spl = splrep(uniformMelFreq, mY)
            mY = splev(binFreqsMel, spl)
        pY = 2 * np.pi * np.random.rand(hN)  # generate phase random values
        Y = np.zeros(N, dtype=complex)  # initialize synthesis spectrum
        Y[:hN] = 10 ** (mY / 20) * np.exp(1j * pY)  # generate positive freq.
        Y[hN:] = 10 ** (mY[-2:0:-1] / 20) * np.exp(-1j * pY[-2:0:-1])  # generate negative freq.
        fftbuffer = np.real(ifft(Y))  # inverse FFT
        y[pout:pout + N] += ws * fftbuffer  # overlap-add
        pout += H
    y = np.delete(y, range(No2))  # delete half of first window
    y = np.delete(y, range(y.size - No2, y.size))  # delete half of the last window
    return y


def stochasticModel(x, H, N, stocf, fs=44100, melScale=1):
    """
	Stochastic analysis/synthesis of a sound, one frame at a time
	x: input array sound, H: hop size, N: fft size 
	stocf: decimation factor of mag spectrum for stochastic analysis, bigger than 0, maximum of 1
	fs: sampling rate
	melScale: choose between linear scale, 0, or mel scale, 1 (should match the analysis)
	returns y: output sound
	"""
    hN = N // 2 + 1  # positive size of fft
    No2 = N // 2  # half of N
    if (hN * stocf < 3):  # raise exception if decimation factor too small
        raise ValueError("Stochastic decimation factor too small")

    if (stocf > 1):  # raise exception if decimation factor too big
        raise ValueError("Stochastic decimation factor above 1")

    if (H <= 0):  # raise error if hop size 0 or negative
        raise ValueError("Hop size (H) smaller or equal to 0")

    if not (UF.isPower2(N)):  # raise error if N not a power of twou
        raise ValueError("FFT size (N) is not a power of 2")

    w = hann(N)  # analysis/synthesis window
    x = np.append(np.zeros(No2), x)  # add zeros at beginning to center first window at sample 0
    x = np.append(x, np.zeros(No2))  # add zeros at the end to analyze last sample
    pin = No2  # initialize sound pointer in middle of analysis window
    pend = x.size - No2  # last sample to start a frame
    y = np.zeros(x.size)  # initialize output array
    if melScale == 1:
        binFreqsMel = hertz_to_mel(np.arange(hN)*fs/float(N))
        uniformMelFreq = np.linspace(binFreqsMel[0], binFreqsMel[-1], hN)
    while pin <= pend:
        # -----analysis-----
        xw = x[pin - No2:pin + No2] * w  # window the input sound
        X = fft(xw)  # compute FFT
        mX = 20 * np.log10(abs(X[:hN]))  # magnitude spectrum of positive frequencies
        if melScale == 1:
            spl = splrep(binFreqsMel, np.maximum(-200, mX))
            stocEnv = resample(splev(uniformMelFreq, spl), int(stocf * hN))  # decimate the mag spectrum
        else:
            stocEnv = resample(np.maximum(-200, mX), int(stocf * hN))  # decimate the mag spectrum
        # -----synthesis-----
        mY = resample(stocEnv, hN)  # interpolate to original size
        if melScale == 1:
            spl = splrep(uniformMelFreq, mY)
            mY = splev(binFreqsMel, spl)
        pY = 2 * np.pi * np.random.rand(hN)  # generate phase random values
        Y = np.zeros(N, dtype=complex)
        Y[:hN] = 10 ** (mY / 20) * np.exp(1j * pY)  # generate positive freq.
        Y[hN:] = 10 ** (mY[-2:0:-1] / 20) * np.exp(-1j * pY[-2:0:-1])  # generate negative freq.
        fftbuffer = np.real(ifft(Y))  # inverse FFT
        y[pin - No2:pin + No2] += w * fftbuffer  # overlap-add
        pin += H  # advance sound pointer
    y = np.delete(y, range(No2))  # delete half of first window which was added
    y = np.delete(y, range(y.size - No2, y.size))  # delete half of last window which was added
    return y
